import os
import threading

import nvidia_smi
import subprocess
import concurrent.futures
import requests
from paho.mqtt import client as mqtt_client
import json
from time import time
from time import sleep
import random
import socket

VERSION = "1.0"

nvidia_smi.nvmlInit()

level_filter_one = 2
debug = False

USERNAME = os.environ["MQTT_USERNAME"]
PASSWORD = os.environ["MQTT_PASSWORD"]
cuda_path = os.environ["CUDA_PATH"]
cado_master = os.environ["CADO_MASTER"]
#"/home/miner/fact_dist/gpu-filter/cuda-ecm"

# --------- MQTT ----------
BROKER = 'fact2.zapto.org'
PORT = 2883
FILTER_REQUEST_TOPIC = "gpu/filter"
FILTER_RESPONSE_TOPIC = "gpu/filter/finished"
FILTER_REGISTER_TOPIC = "gpu/filter/register"
FILTER_SETTINGS_TOPIC = "gpu/filter/settings"

FIRST_RECONNECT_DELAY = 1
RECONNECT_RATE = 2
MAX_RECONNECT_DELAY = 60

client = mqtt_client.Client('gpu_filter-' + USERNAME + "_" + str(random.randint(0, 1000)))
# --------- MQTT ----------

settings = {}
offset_position = 0
chunk_size = 150
hostname = socket.gethostname()

def create_config_file(level, idx, filename, filter_level):
    if filter_level == 1:
        ecmparams = {"b1": [11000, 50000, 100000, 3000000, 11000000, 43000000],
                     "curves": [107, 400, 910, 2351, 4482, 7557]}

    conf_data = """

; Example config file for co-ecm
[general]

; server or file
mode = file

; Logfile location
logfile = ./log.txt

; Output file of abandoned, i.e. unsolved tasks.
; Format is the same as the output format, without listing factors
;
; Example line:
; 44800523911798220433379600867; # effort 112 
output_abandoned = ./abandoned.txt

; Log level
;
;   1: "VERBOSE",
;   2: "DEBUG",
;   3: "INFO",
;   4: "WARNING",
;   5: "ERROR",
;   6: "FATAL",
;   7: "NONE"
; Default is set at compile time.
loglevel = 3    

; Use a random seed for the random number generator used to generate points and 
; curves. If set to 'false', each run of the program will behave the same
; provided the same input data.
; Default: true
random = true

[server]
port = 11111

[file]
; Input file.
; The input file should contain a single number to be factored per line. Lines
; starting with anything but a digit are skipped.
;
; Example line:
; 44800523911798220433379600867
input = ./input{}.txt


; Output file.
; Each fully factored input number is appended to the output on its own line in
; the format 
; (input number);(factor),(factor),(factor), # effort: (number of curves)
;
; Example line:
; 44800523911798220433379600867;224536506062699,199524454608233, # effort: 12
output = ./output{}.txt


[cuda]

; Number of concurrent cuda streams to issue to GPU 
; Default: 2
streams = 1 

; Number of threads per block for cuda kernel launches.
; Set to auto to determine setting for maximum parallel resident blocks per SM at runtime.
; Note: The settings determined by 'auto' are not always automatically the optimal setting for maximum throughput.
; Default: auto
threads_per_block = auto 

; Constant memory is used for (smaller) scalars during point multiplication.
; When the scalar is too large to fit into constant memory or this option is set 
; to 'false', global device memory is used.
; Default: true
use_const_memory = true

[ecm]
; Redo ECM until numbers are fully factored.
; If set to false, only the first factor is returned.
; Default: false
;find_all_factors = true
find_all_factors = false

; Set the computation of the scalar s for point multiplication. With
; 'powersmooth' set to 'true', then s = lcm(2, ..., b1). If set to false,
; s = primorial(2, ..., b1), i.e. the product of all primes less than or equal
; to b1. 
; Default: true
powersmooth = true

b1 = {}
b2 = 100000


; Maximum effort per input number.
; With each curve, the already spent effort is incremented. Thus, with effort
; set to 100, ecm stage1 (and stage2) will be executed on 100 curves per input
; number.
; Default: 10
effort = {} 

; Set the curve generator function. 
; Use 2 under normal circumstances.
;   0: "Naive"
;   1: "GKL2016_j1"
;   2: "GKL2016_j4"
; Default: 2
curve_gen = 2   

; Use only points for finding factors that are off curve.
; After point multiplication, use all resulting points to find factors. If set 
; to 'false' coordinates of points will be checked that do not fulfill the curve
; equation.
; Settings for stage1 and stage2 respectively.
; Default: true
stage1.check_all = false
stage2.check_all = true

; Enable/Disable stage 2.
; If set to 'false', only stage 1 of ECM is performed.
; Default: true
stage2.enabled = false

; Set the window size for stage 2
; Default: 2310
;stage2.window_size = 2310
""".format(idx, idx, ecmparams["b1"][level], ecmparams["curves"][level])

    with open(filename, 'w') as inputfile:
        inputfile.write(conf_data)
        inputfile.close()


def submit_unfiltered_chunk(current_block, candidates, segment):
    try:
        print("submitting unfiltered chunk... for block: ", current_block)
        payload = {'height': current_block, 'slice': candidates, "segment": segment}
        url = "http://" + cado_master + "/submitunfilteredchunk"
        x = requests.post(url, json=payload)
    except:
        print("failed to submit candidate to cado master. will continue ...")

def report_candidates(height, level, segment, size, type):
    url = 'http://fact.zapto.org/candidates/submitstats/v2'

    try:
        payload = {'height': height, 'level': level, 'segment': segment, 'size': size, 'type': type}
        x = requests.post(url, json=payload)
    except:
        print("failed to submit candidate stats....")

def gpu_sieving(level, gpu_idx, config, candidates, nBits):
    # Clean input/output files
    subprocess.run("rm -rf output" + str(gpu_idx) + ".txt", capture_output=True, shell=True)
    subprocess.run("rm -rf input" + str(gpu_idx) + ".txt", capture_output=True, shell=True)

    # Create input file for GPU work
    with open("input" + str(gpu_idx) + ".txt", 'w') as inputfile:
        inputfile.write("\n".join([str(idx) + " " + str(n) for idx, n in enumerate(candidates)]))
        inputfile.close()

    # Create a copy of the config file for this running instance
    output = "output" + str(gpu_idx) + ".txt"
    config_name = "gpu_config_" + str(level) + "_" + str(gpu_idx) + ".txt"

    start = time()
    parse = subprocess.run("CUDA_VISIBLE_DEVICES=" + str(gpu_idx) + " " + cuda_path + " -c " + config_name, capture_output=True, shell=True)

    if parse.returncode != 0:
        return [-1]

    survivors = []
    with (open(output, 'r') as outputfile):
        read = [line for line in outputfile if "DONE" not in line]
        survivor_index = [int(g[0]) for line in read if (g := line.split())[1] == "1"]

        if len(candidates) != 0:
            survivors = [candidates[index] for index in survivor_index]
            solution_index = [(int(g[0]), int(g[1])) for line in read if
                              len(g := line.split()) == 3 and int(g[1]).bit_length() == (nBits // 2 + (nBits & 1))]
            solutions = [candidates[index[0]] for index in solution_index]

            # Check if we got lucky and found a solution
            for i_p, n in zip(solution_index, solutions):
                tp, tq = int(i_p[1]), int(n) // int(i_p[1])
                p = min(tp, tq)
                q = max(tp, tq)

                if p.bit_length() == (nBits // 2 + (nBits & 1)):
                    # TODO -> submit solution
                    print("solution found in sieving phase!")


    print("GPU ID " + str(gpu_idx) + " Total Time: ", time() - start, " Seconds.   Survivors: ", len(survivors))

    return survivors

def filter(candidates, nBits):
    global chunk_size
    global offset_position
    #############################################################################################
    #                                       GPU Filtering                                       #
    #############################################################################################
    deviceCount = nvidia_smi.nvmlDeviceGetCount()

    print("gpu devices: ", deviceCount)

    if 'chunkSize' in settings:
        chunk_size = settings["chunkSize"]
    if 'offsetPosition' in settings:
        offset_position = settings["offsetPosition"]

    print("offset position: ", offset_position, " chunk_size:", chunk_size)

    start_idx = offset_position * chunk_size
    end_idx = start_idx + chunk_size
    if end_idx > len(candidates):
        end_idx = len(candidates)

    print("processing candidates from idx: ", start_idx, " to idx:", end_idx)

    candidates = candidates[start_idx:end_idx]

    print("took", len(candidates), " candidates")

    # Clean input/output files
    subprocess.run("rm -rf output*.txt", capture_output=True, shell=True)
    subprocess.run("rm -rf input*.txt", capture_output=True, shell=True)

    for level in range(level_filter_one):
        for idx in range(deviceCount):
            config_name = "gpu_config_" + str(level) + "_" + str(idx) + ".txt"
            create_config_file(level, idx, config_name, 1)

    for level in range(level_filter_one):
        cand_size = len(candidates) // deviceCount
        gpu_inputs = [(level, gpu_idx, "gpu_config_" + str(level) + "_" + str(gpu_idx) + ".txt",
                       candidates[gpu_idx * cand_size: min((gpu_idx + 1) * cand_size, len(candidates))], nBits) for
                      gpu_idx in range(deviceCount)]

        with concurrent.futures.ThreadPoolExecutor(max_workers=deviceCount) as executor:
            future = [executor.submit(gpu_sieving, *indata) for indata in gpu_inputs]
            return_value = [future[idx].result() for idx in range(len(future))]
            return_value = [x for xs in return_value for x in xs]
            if any([a == -1 for a in return_value]):
                return []
            candidates = return_value

    print("returning: ", len(candidates), " survivors")
    return candidates

def on_connect(client, userdata, flags, rc):
    if rc == 0 and client.is_connected():
        print("Connected to MQTT Broker!")
        client.subscribe(FILTER_REQUEST_TOPIC)
        client.subscribe(FILTER_SETTINGS_TOPIC)

        send_register_payload()
    else:
        print('Failed to connect, return code ', rc)

def publish(payload, topic):
    result = client.publish(topic, payload)
    # result: [0, 1]
    status = result[0]
    if status != 0:
        print('Failed to send message to topic:', topic)

def on_disconnect(client, userdata, rc):
    print("Disconnected with result code: ", rc)
    reconnect_count, reconnect_delay = 0, FIRST_RECONNECT_DELAY
    while True:
        print("Reconnecting in ", reconnect_delay, "... ")
        sleep(reconnect_delay)

        try:
            client.reconnect()
            print("Reconnected successfully!")
            return
        except Exception as err:
            print("Reconnect failed. Retrying... Error was:", err)

        reconnect_delay *= RECONNECT_RATE
        reconnect_delay = min(reconnect_delay, MAX_RECONNECT_DELAY)
        reconnect_count += 1

def on_message(client, userdata, msg):
    global settings
    global hostname

    payload = msg.payload.decode()
    received_json = json.loads(payload)

    if msg.topic == FILTER_SETTINGS_TOPIC and debug is True:
        print("settings: ", received_json)

    if msg.topic == FILTER_SETTINGS_TOPIC and received_json["hostname"] == hostname:
        print("settings: ", received_json)
        settings = received_json
    #print(received_json)

    if msg.topic == FILTER_REQUEST_TOPIC:
        nbits = received_json["nbits"]
        segment = received_json["segment"]
        candidates = received_json["cands"]
        height = received_json["height"]

        print ("filter request. segment: ", segment, "height: ", height, "len: ", len(candidates))

        if "segment" not in settings:
            print("settings has no segment configured")
            return
        if int(settings["segment"]) != int(segment):
            print("settings segment: ", settings["segment"], " not matching ", segment)
            return

        print ("processing ... ")

        survivors = filter(candidates, nbits)
        response = {"survivors": survivors, "segment": segment, "height": height}

        report_candidates(height, 1, segment, len(survivors), "generated")
        submit_unfiltered_chunk(height, survivors, int(segment))
        publish(json.dumps(response), FILTER_RESPONSE_TOPIC)

def send_register_payload():
    register_payload = json.dumps({'hostname': hostname})
    publish(register_payload, FILTER_REGISTER_TOPIC)

def register_thread():
    while(True):
        if client.is_connected():
            print("registering ...")
            send_register_payload()
        sleep(30)

def connect_mqtt():
    client.username_pw_set(USERNAME, PASSWORD)
    client.on_connect = on_connect
    client.on_message = on_message
    client.connect(BROKER, PORT, keepalive=120)
    client.on_disconnect = on_disconnect
    client.loop_forever()

print("init version", VERSION, "with hostname:", hostname)

t = threading.Thread(target=register_thread, args=())
t.start()
connect_mqtt()
